#include "local_search.h"

#include "plugin.h"

using namespace std;

template<class Entry>
LocalSearch<Entry>::LocalSearch(const Options &opts)
    : SearchEngine(opts),
      global_search(opts.get<SearchEngine *>("global_search")),
      local_search_pattern(opts.get<SearchEngine *>("local_search")),
      local_search_instance(0),
      global_min_h(numeric_limits<int>::max()),
      local_tries(0),
      evaluated(0),
      max_num_stalled(opts.get<int>("max_stalled")),
      max_number_of_local_tries(opts.get<int>("max_local_tries")),
      local_search_size(opts.get<int>("local_search_size")) {
}

template<class Entry>
LocalSearch<Entry>::~LocalSearch() {
}

template<class Entry>
void LocalSearch<Entry>::initialize() {
    global_search->initialize();
    cout << "initialize: get open list" << endl;
    open_list = global_search->get_options().get<OpenList<Entry> * >("open");
    open_list->get_involved_heuristics(heuristics);
    local_open_list = local_search_pattern->get_options().get<OpenList<Entry> * >("open");
}

template<class Entry>
SearchStatus LocalSearch<Entry>::step() {
    SearchStatus status = FAILED;
    if (!local_search_instance) {
        status = global_search->step();
        if (status == IN_PROGRESS) {
            SearchProgress &search_progress = global_search->search_progress;
            if (global_min_h > search_progress.get_min_h()) {
                global_min_h = search_progress.get_min_h();
                local_tries = 0;
                reset_num_stalled();
            } else if (get_num_stalled() >= max_num_stalled && local_tries < max_number_of_local_tries) {
                if (initialize_local_search() == SOLVED) {
                    return SOLVED;
                }
            }
        }
    } else {
        status = local_step();
    }
    return status;
}

template<class Entry>
SearchStatus LocalSearch<Entry>::initialize_local_search() {
    SearchSpace &search_space = global_search->search_space;
    Options opts = global_search->get_options();

    cout << "Status: Stalled: " << get_num_stalled() << " Tries: " << local_tries << endl;
    Entry entry = open_list->remove_min();

    // Some searches mignt not have tested it yet and then skip this state.
    if (global_search->check_goal_and_set_plan(g_state_registry->lookup_state(get_stateid_from_entry(entry)))) {
        return SOLVED;
    }

    opts.set<Entry>("root", entry);
    assert(open_list->empty());
    opts.set<OpenList<Entry> *>("open", local_open_list);
    opts.set<SearchSpace *>("search_space", &search_space);


    local_search_instance = local_search_pattern->new_instance(opts);
    local_search_instance->initialize();
    local_step_number = 0;

    cout << "Local Search current h_min: " << global_min_h << " local_h_min: " << local_search_instance->search_progress.get_min_h() << endl;
    return IN_PROGRESS;
}

template<class Entry>
SearchStatus LocalSearch<Entry>::local_step() {
    SearchStatus status = IN_PROGRESS;
    status = local_search_instance->step();
    ++local_step_number;
    //        cout << "h_min: " << h_min << " local_h_min: " << local_search->search_progress.get_global_h_min() << endl;

    if (local_step_number >= local_search_size ||
        status != IN_PROGRESS ||
        global_min_h > local_search_instance->search_progress.get_min_h()
        ) {
        int added = merge_openlist();
        cout << "Merged " << added << " entries from the local search." << endl;
        reset_num_stalled();
        ++local_tries;

        delete local_search_instance;
        local_search_instance = 0;

        // The local search failed but this is not a reason to report the whole search as failed.
        if (status == FAILED) {
            status = IN_PROGRESS;
        }
    }
    return status;
}

template<class Entry>
int LocalSearch<Entry>::merge_openlist() {
    int added = 0;

    while (!local_open_list->empty()) {
        Entry entry = local_open_list->remove_min();
        SearchNode node = global_search->search_space.get_node(g_state_registry->lookup_state(get_stateid_from_entry(entry)));
        for (set<Heuristic *>::const_iterator h = heuristics.begin(); h != heuristics.end(); ++h) {
            (*h)->evaluate(node.get_state());
        }
        open_list->evaluate(node.get_g(), false);
        open_list->insert(entry);
        ++added;
        //        if(h_min != node.get_h()) {
        //            cout << "added " << *it << " with h: " << node.get_h() << " is open: " << node.is_open() << " is new: " << node.is_new() << endl;
        //        }
    }
    return added;
}

template<class Entry>
void LocalSearch<Entry>::statistics() const {
    global_search->statistics();
}

template<class Entry>
bool LocalSearch<Entry>::found_solution() const {
    return global_search->found_solution();
}

template<class Entry>
const SearchEngine::Plan &LocalSearch<Entry>::get_plan() const {
    return global_search->get_plan();
}

template<>
StateID LocalSearch<StateID>::get_stateid_from_entry(const StateID &entry) {
    return entry;
}

template<>
StateID LocalSearch<OpenListEntryLazy>::get_stateid_from_entry(const OpenListEntryLazy &entry) {
    return entry.first;
}

template<class Entry>
static SearchEngine *_parse(OptionParser &parser) {
    parser.document_synopsis("Best-first search with local exploration",
                             "Best-first search with local exploration, if the search detects a stall, "
                             "a local search is started on the next node with the best heuristic value."
                             "The local search engine can either be a seach engine of the same type as "
                             "the global search engine, or a random search.");
    parser.add_option<SearchEngine *>("global_search", "Search engine used for the global search");
    parser.add_option<SearchEngine *>("local_search", "Search engine used for the local search");
    parser.add_option<int>("max_stalled", "Number of salled expansions before a local search is started.", "1000");
    parser.add_option<int>("max_local_tries", "Maximum number of local searchs with out improvement.", "100");
    parser.add_option<int>("local_search_size", "Number of steps for the local search", "1000");
    SearchEngine::add_options_to_parser(parser);
    Options opts = parser.parse();

    if (!parser.dry_run()) {
        return new LocalSearch<Entry>(opts);
    }
    return 0;
}

static Plugin<SearchEngine> _plugin_eager("eager_local", _parse<StateID>);

static Plugin<SearchEngine> _plugin_lazy("lazy_local", _parse<OpenListEntryLazy>);
