#ifndef LOCAL_SEARCH_H
#define LOCAL_SEARCH_H

#include "global_state.h"
#include "scalar_evaluator.h"
#include "search_engine.h"
#include "search_progress.h"
#include "search_space.h"
#include "option_parser_util.h"

#include "lazy_search.h"

#include "open_lists/open_list.h"

#include <vector>
#include <limits>


/*
  Best-first search with local exploration, if the search detects a stall,
  a local search is started on the next node with the best heuristic value.
  The local search engine can either be a seach engine of the same type as
  the global search engine, or a random search.
  Based on:
    Fan Xie, Martin Mueller, and Robert C. Holte. Adding local exploration to
    greedy best-first search in satisficing planning. In Proceedings of the
    Twenty-Eighth AAAI Conference on Artificial Intelligence (AAAI 2014). AAAI Press, 2014.
*/

template<class Entry>
class LocalSearch : public SearchEngine {
    SearchEngine *global_search;
    const SearchEngine *local_search_pattern;
    SearchEngine *local_search_instance;

    int global_min_h;
    int local_tries;
    int local_step_number;

    int evaluated;

    int max_num_stalled;
    int max_number_of_local_tries;
    int local_search_size;

    OpenList<Entry> *open_list;
    OpenList<Entry> *local_open_list;
    std::set<Heuristic *> heuristics;

    void reset_num_stalled() {
        evaluated = global_search->search_progress.get_evaluated_states();
    }

    int get_num_stalled() {
        return global_search->search_progress.get_evaluated_states() - evaluated;
    }

protected:
    virtual void initialize();

    virtual SearchStatus step();

    int merge_openlist();

    StateID get_stateid_from_entry(const Entry &entry);

    SearchStatus initialize_local_search();
    SearchStatus local_step();
public:

    explicit LocalSearch(const Options &opts);
    virtual ~LocalSearch();
    virtual void statistics() const;
    virtual bool found_solution() const;
    virtual const Plan &get_plan() const;
};
#endif
