#ifndef OPEN_LISTS_DIVERSE_OPEN_LIST_H
#define OPEN_LISTS_DIVERSE_OPEN_LIST_H

#include "open_list.h"
#include "../option_parser.h"
#include "../state_id.h"

#include <unordered_map>
#include <vector>

class GlobalOperator;
class Heuristic;

/*
  Diverse best first search diversifies search by probabalistic selecting nodes
  which not have the best heuristic value.
  Based on:
    Tatsuya Imai and Akihiro Kishimoto. A novel technique for avoiding plateaus
    of greedy best-first search in satisficing planning. In Wolfram Burgard and Dan Roth,
    editors, Proceedings of the Twenty-Fifth AAAI Conference on Artificial Intelligence (AAAI 2011),
    pages 985-991. AAAI Press, 2011
*/

template<class Entry>
class DiverseOpenList : public OpenList<Entry>{
    int last_g_value;
    std::unordered_map<StateID, int, __gnu_cxx::hash<StateID> > state_id_to_g;

    void merge_open_lists();
    void reset_local_open_list();

    StateID get_stateid_from_entry(const Entry &entry);
protected:
    OpenList<Entry> *global_open_list;
    OpenList<Entry> *local_open_list;

    std::vector<Heuristic *> global_heuristics;
    std::vector<Heuristic *> local_heuristics;

    int num_remaining_steps;

    virtual Evaluator *get_evaluator() {return this; }

public:

    explicit DiverseOpenList(const Options &opts);
    virtual ~DiverseOpenList();

    virtual int insert(const Entry &entry);
    virtual Entry remove_min(std::vector<int> *key);
    virtual bool empty() const;
    virtual void clear();

    virtual void evaluate(int g, bool preferred);
    virtual bool is_dead_end() const;
    virtual bool dead_end_is_reliable() const;
    virtual void get_involved_heuristics(std::set<Heuristic *> &hset);

    static OpenList<Entry> *_parse(OptionParser &parser);
};

#include "diverse_open_list.cc"

// HACK! Need a better strategy of dealing with templates, also in the Makefile.

#endif
