#ifndef OPEN_LISTS_RANDOM_OPEN_LIST_H
#define OPEN_LISTS_RANDOM_OPEN_LIST_H

#include "open_list.h"

/*
  Random open list based on

    Richard Valenzano, Nathan R. Sturtevant, Jonathan Schaeffer and Fan Xie.
    A Comparison of Knowledge-Based GBFS Enhancements and Knowledge-Free Exploration.
    In Proceedings of the Twenty-Fourth International Conference on Automated
    Planning and Scheduling (ICAPS 2014), pages 375-379. AAAI Press, 2014.

  With probability epsilon the next entry is selected uniformly randomly,
  otherwise the minimum entry is chosen. While the original
  implementation by Valenzano et al. is based on buckets (personal
  communication), this implementation uses a heap. If n is the number of
  entries and m is the number of buckets, the worst-case times of both
  implementations are:

                          Buckets     Heap
  Insert entry            O(1)        O(log(n))
  Remove random entry     O(m)        O(log(n))
  Remove minimum entry    O(1)        O(log(n))

  An experimental evaluation of both implementations showed that the
  heap-based version is faster in practice. To obtain FIFO tiebreaking
  as in the bucket-based implementation we assign increasing IDs to new entries.
*/

typedef struct {
    int id;
    int h;
} RandomHeapKey;

template<class Entry>
class RandomOpenList : public OpenList<Entry>{
    struct GreaterThanComparator {
        bool operator()(const pair<RandomHeapKey, Entry> &a,
                        const pair<RandomHeapKey, Entry> &b) const;
    };

    GreaterThanComparator comparator;

    std::vector<std::pair<RandomHeapKey, Entry> > open_list;
    ScalarEvaluator *evaluator;

    int last_evaluated_value;
    int last_preferred;
    bool dead_end;
    bool dead_end_reliable;

    double epsilon;
    int size;
    int next_id;

protected:
    ScalarEvaluator *get_evaluator() {return evaluator; }

public:
    explicit RandomOpenList(const Options &opts);
    virtual ~RandomOpenList();

    virtual int insert(const Entry &entry);
    virtual Entry remove_min(std::vector<int> *key = 0);

    virtual bool empty() const;
    virtual void clear();

    virtual void evaluate(int g, bool preferred);
    virtual bool is_dead_end() const;
    virtual bool dead_end_is_reliable() const;
    virtual void get_involved_heuristics(std::set<Heuristic *> &hset);

    static OpenList<Entry> *_parse(OptionParser &parser);
};

#include "random_open_list.cc"

// HACK! Need a better strategy of dealing with templates, also in the Makefile.
#endif
