#include "random_walk_search.h"

#include "globals.h"
#include "heuristic.h"
#include "option_parser.h"
#include "plugin.h"
#include "rng.h"
#include "successor_generator.h"

using namespace std;

RandomWalkSearch::RandomWalkSearch(const Options &opts)
    : SearchEngine(opts),
      walk_type(opts.get<RandomWalkType *>("walk_type")),
      max_explorations_without_improvement(opts.get<int>("max_explorations_without_improvement")),
      walks_per_exploration(opts.get<int>("walks_per_exploration")),
      default_walk_length(opts.get<int>("default_walk_length")),
      heuristic(opts.get<Heuristic *>("heuristic")),
      extending_period(opts.get<int>("extending_period")),
      extending_rate(opts.get<double>("extending_rate")),
      weight_for_acceptable_progress(opts.contains("weight_for_acceptable_progress") ? opts.get<double>("weight_for_acceptable_progress") : -1),
      use_acceptable_progress(opts.contains("weight_for_acceptable_progress")),
      last_acceptable_progress(0),
      current_state(g_initial_state()),
      h_min(numeric_limits<int>::max()),
      init_h(h_min),
      num_explorations_without_improvement(0),
      num_resets(0) {
}

RandomWalkSearch::~RandomWalkSearch() {
}

void RandomWalkSearch::initialize() {
    search_progress.add_heuristic(heuristic);
    heuristic->evaluate(current_state);
    search_progress.inc_evaluations();
    search_progress.inc_evaluated_states();
    init_h = heuristic->get_heuristic();

    reset();
    search_progress.get_initial_h_values();
    search_progress.check_h_progress(0);
}

void RandomWalkSearch::build_plan_in_search_space() {
    GlobalState parent = g_initial_state();
    SearchNode root = search_space.get_node(parent);
    root.open_initial(0);

    for (Path::iterator it = global_path.begin(); it != global_path.end(); ++it) {
        const GlobalState &tmp_state = it->first;
        const GlobalOperator *op = it->second;
        SearchNode node = search_space.get_node(tmp_state);
        node.open(0, search_space.get_node(parent), op);
        parent = tmp_state;
    }
    assert(test_goal(parent));
}

void RandomWalkSearch::reset() {
    num_explorations_without_improvement = 0;
    current_state = g_initial_state();
    h_min = init_h;

    visited.clear();
    global_path.clear();
    walk_type->reset();
    ++num_resets;
}

void RandomWalkSearch::statistics() const {
    search_progress.print_statistics();
    search_space.statistics();
    cout << "Number of resets: " << num_resets << endl;
}

RandomWalkSearch::WalkResult RandomWalkSearch::do_single_monte_carlo_random_walk(
    const GlobalState &root, Path &tmp_path, const int walk_length) {
    assert(tmp_path.empty());
    GlobalState tmp_state = root;
    for (int j = 0; j < walk_length; ++j) {
        vector<const GlobalOperator *> applicable_ops;
        g_successor_generator->generate_applicable_ops(tmp_state, applicable_ops);
        if (applicable_ops.empty()) {
            return DEAD_END;
        }

        const GlobalOperator *op = walk_type->choose_operator(applicable_ops);
        tmp_state = g_state_registry->get_successor_state(tmp_state, *op);
        tmp_path.push_back(make_pair(tmp_state, op));

        if (test_goal(tmp_state)) {
            return GOAL;
        }
    }
    return NODE;
}

RandomWalkSearch::Path RandomWalkSearch::do_monte_carlo_exploration(const GlobalState &root) {
    Path accepted_path;
    int local_h_min = numeric_limits<int>::max();

    int walk_length = default_walk_length;
    int walks_without_improvement = 0;

    for (int i = 0; i < walks_per_exploration; ++i) {
        Path tmp_path;
        tmp_path.reserve(walk_length);

        WalkResult walk_result = do_single_monte_carlo_random_walk(root, tmp_path, walk_length);
        if (walk_result != RandomWalkSearch::DEAD_END) {
            GlobalState &walk_end_state = tmp_path.back().first;
            heuristic->evaluate(walk_end_state);
            search_progress.inc_evaluated_states();
            search_progress.inc_evaluations();
            if (heuristic->is_dead_end()) {
                walk_result = RandomWalkSearch::DEAD_END;
            }
        }

        walk_type->handle_walk(tmp_path, walk_result);

        if (walk_result == RandomWalkSearch::GOAL) {
            return tmp_path;
        } else if (walk_result == RandomWalkSearch::DEAD_END) {
            search_progress.inc_dead_ends();
            ++walks_without_improvement;
        } else {
            int walk_end_state_h = heuristic->get_heuristic();
            if (walk_end_state_h < local_h_min) {
                swap(accepted_path, tmp_path);
                local_h_min = walk_end_state_h;
                walks_without_improvement = 0;
                if (use_acceptable_progress) {
                    int progress = max(0, h_min - local_h_min);
                    double acceptable_progress = compute_acceptable_progress(progress);
                    if (progress >= acceptable_progress) {
                        return accepted_path;
                    }
                }
            } else if (++walks_without_improvement > extending_period) {
                walk_length *= extending_rate;
                walks_without_improvement = 0;
            }
        }
    }
    return accepted_path;
}

void RandomWalkSearch::update_global_path(const Path &path) {
    for (Path::const_iterator it = path.begin(); it != path.end(); ++it) {
        StateID state_id = it->first.get_id();
        if (visited.find(state_id) == visited.end()) {
            visited.insert(state_id);
            global_path.push_back(*it);
        } else {
            while (global_path.back().first.get_id() != state_id) {
                visited.erase(global_path.back().first.get_id());
                global_path.pop_back();
            }
        }
    }
}

SearchStatus RandomWalkSearch::step() {
    if (test_goal(current_state)) {
        build_plan_in_search_space();
        check_goal_and_set_plan(current_state);
        return SOLVED;
    }

    Path accepted_path = do_monte_carlo_exploration(current_state);
    if (!accepted_path.empty()) {
        GlobalState &exploration_end_state = accepted_path.back().first;
        heuristic->evaluate(exploration_end_state);
        search_progress.inc_evaluations();
        int exploration_end_state_h = heuristic->get_heuristic();

        if (exploration_end_state_h < h_min) {
            current_state = exploration_end_state;
            update_global_path(accepted_path);

            if (use_acceptable_progress) {
                last_acceptable_progress = compute_acceptable_progress(h_min - exploration_end_state_h);
            }

            walk_type->handle_progress(heuristic);

            h_min = exploration_end_state_h;
            search_progress.check_h_progress(0);

            num_explorations_without_improvement = 0;
            return IN_PROGRESS;
        }
    }
    if (++num_explorations_without_improvement > max_explorations_without_improvement) {
        reset();
    }
    return IN_PROGRESS;
}

ProbabilisticWalk::ProbabilisticWalk(double tau)
    : tau(tau),
      q_values(g_operators.size(), 0) {
}

ProbabilisticWalk::~ProbabilisticWalk() {
}

void ProbabilisticWalk::reset() {
    fill(q_values.begin(), q_values.end(), 0);
}

const GlobalOperator *ProbabilisticWalk::choose_operator(const std::vector<const GlobalOperator *> &applicable_ops) {
    double sum = 0;
    const GlobalOperator *out = 0;
    for (size_t pos = 0; pos < applicable_ops.size(); ++pos) {
        const GlobalOperator *op = applicable_ops[pos];
        double prob = exp(q_values[op->get_id()] / tau);
        sum += prob;
        if (g_rng() * sum < prob) {
            out = op;
        }
    }
    assert(out);
    return out;
}

RandomWalkType::RandomWalkType() {
}

RandomWalkType::~RandomWalkType() {
}


PureWalk::PureWalk() {
    cout << "Using PureWalk" << endl;
}

PureWalk::~PureWalk() {
}

const GlobalOperator *PureWalk::choose_operator(const std::vector<const GlobalOperator *> &applicable_ops) {
    return *g_rng.choose(applicable_ops);
}

MDAWalk::MDAWalk(const Options &opts)
    : ProbabilisticWalk(opts.get<double>("tau")),
      num_operator_used_in_failed_walks(g_operators.size(), 0),
      num_operator_used_in_successful_walks(g_operators.size(), 0) {
    cout << "Using MDAWalk" << endl;
}

void MDAWalk::reset() {
    ProbabilisticWalk::reset();
    fill(num_operator_used_in_successful_walks.begin(), num_operator_used_in_successful_walks.end(), 0);
    fill(num_operator_used_in_failed_walks.begin(), num_operator_used_in_failed_walks.end(), 0);
}

void MDAWalk::handle_walk(const RandomWalkSearch::Path &path, RandomWalkSearch::WalkResult walk_result) {
    for (size_t i = 0; i < path.size(); ++i) {
        size_t op_id = path[i].second->get_id();
        int &num_failed = num_operator_used_in_failed_walks[op_id];
        int &num_success = num_operator_used_in_successful_walks[op_id];
        if (walk_result == RandomWalkSearch::DEAD_END) {
            ++num_failed;
        } else {
            ++num_success;
        }
        double sum = num_failed + num_success;
        assert(sum >= 0); // overflow
        q_values[op_id] = (sum == 0) ? 0 : -num_failed / sum;
    }
}

MHAWalk::MHAWalk(const Options &opts)
    : ProbabilisticWalk(opts.get<double>("tau")) {
    cout << "Using MHAWalk" << endl;
}

void MHAWalk::handle_progress(Heuristic *heuristic) {
    vector<const GlobalOperator *> helpful;
    heuristic->get_preferred_operators(helpful);
    for (size_t i = 0; i < helpful.size(); ++i) {
        ++q_values[ helpful[i]->get_id() ];
    }
}


static RandomWalkType *_parse_pure_walk(OptionParser &parser) {
    parser.document_synopsis("Pure random walk.",
                             "Pure random walk without helpful actions or dead end avoidance.");
    if (parser.dry_run())
        return 0;
    else
        return new PureWalk();
}

static Plugin<RandomWalkType> _plugin_pure_walk("pure", _parse_pure_walk);


static RandomWalkType *_parse_mha_walk(OptionParser &parser) {
    parser.document_synopsis("Monte-Carlo helpful actions.",
                             "Monte-Carlo helpful actions, preferred operators are "
                             "evaluated each time the heuristic is improved, "
                             "operators that lead to an improved heuristic are "
                             "selected with a higher probability.");
    parser.add_option<double>("tau", "Temperature used for MHA.", "10");
    Options opts = parser.parse();

    if (opts.get<double>("tau") <= 0) {
        parser.error("tau must be greater than 0.");
    }

    if (parser.dry_run())
        return 0;
    else
        return new MHAWalk(opts);
}

static Plugin<RandomWalkType> _plugin_mha_walk("mha", _parse_mha_walk);

static RandomWalkType *_parse_mda_walk(OptionParser &parser) {
    parser.document_synopsis("Monte-Carlo dead end avoidance.",
                             "For each operator we record how often it was part "
                             "of successful and failed walks. Failed walks are "
                             "the ones that end in dead end states, all other "
                             "walks are counted as successful. During the walks"
                             "we select operators that lead to non-dead end "
                             "states with higher probability.");
    parser.add_option<double>("tau", "Temperature used for MDA.", "0.5");
    Options opts = parser.parse();

    if (opts.get<double>("tau") <= 0) {
        parser.error("tau must be greater than 0.");
    }

    if (parser.dry_run())
        return 0;
    else
        return new MDAWalk(opts);
}

static Plugin<RandomWalkType> _plugin_mda_walk("mda", _parse_mda_walk);


static SearchEngine *_parse_random_walks_search(OptionParser &parser) {
    parser.document_synopsis("Monte-Carlo Random walk search",
                             "Search based on random walks. Only the end point "
                             "of a random path is evaluated.");

    parser.add_option<Heuristic *>("heuristic", "The heuristic used for evaluating the last state of each walk.");
    parser.add_option<int>("max_explorations_without_improvement", "Number of maximum tries until the search restarts.", "7");
    parser.add_option<int>("walks_per_exploration", "The number of walks for each random exploration.", "2000");
    parser.add_option<int>("default_walk_length", "The initial length of a walk.", "10");
    parser.add_option<int>("extending_period", "If more walks then defined here "
                           "failed to improve the heuristic value, the default_walk_length "
                           "for the current search step is increased by the factor extending_rate.", "300");
    parser.add_option<double>("extending_rate", "Rate of expansion of the default_walk_length, see extending_period.", "1.5");
    parser.add_option<double>("weight_for_acceptable_progress", "Parameter used for acceptable progress computation.", "", OptionFlags(false));

    parser.add_option<RandomWalkType *>("walk_type", "Defines the Walk Type", "pure");

    SearchEngine::add_options_to_parser(parser);
    Options opts = parser.parse();

    if (opts.get<int>("max_explorations_without_improvement") <= 0) {
        parser.error("max_explorations_without_improvement must be greater than 0.");
    }

    if (opts.get<int>("walks_per_exploration") <= 0) {
        parser.error("walks_per_exploration must be greater than 0.");
    }
    if (opts.get<int>("default_walk_length") <= 0) {
        parser.error("default_walk_length must be greater than 0.");
    }
    if (opts.get<int>("extending_period") <= 0) {
        parser.error("extending_period must be greater than 0.");
    }
    if (opts.get<double>("extending_rate") <= 0) {
        parser.error("extending_rate must be greater than 0.");
    }
    if (opts.contains("weight_for_acceptable_progress") && !in_range(opts.get<double>("weight_for_acceptable_progress"), 0.0, 1.0)) {
        parser.error("weight_for_acceptable_progress must be between 0 and 1.");
    }

    RandomWalkSearch *engine = 0;
    if (!parser.dry_run()) {
        engine = new RandomWalkSearch(opts);
    }

    return engine;
}


static Plugin<SearchEngine> _plugin("random_walk", _parse_random_walks_search);
